# This file was generated automatically by Qleany's generator, edit at your own risk! 
# If you do, be careful to not overwrite it when you run the generator again.

find_package(Qt6 COMPONENTS Core  REQUIRED)

set(LIBRARY_NAME {{ application_name }}-contracts-cqrs-{{ feature.feature_name_spinal }})

set(SRC

    {% for file in files %}
    {{ file }}
    {%- endfor %}
)
qt_add_library(${LIBRARY_NAME} STATIC ${SRC})

target_include_directories(${LIBRARY_NAME} PUBLIC "${CMAKE_CURRENT_SOURCE_DIR}")

target_link_libraries(${LIBRARY_NAME} PRIVATE {{ application_name }}-contracts)
target_link_libraries(${LIBRARY_NAME} PRIVATE {{ application_name }}-domain)
target_link_libraries(${LIBRARY_NAME} PUBLIC {{ application_name }}-contracts-dto-{{ feature.feature_name_spinal }})

target_link_libraries(${LIBRARY_NAME}
    PRIVATE Qt6::Core
)