// This file was generated automatically by Qleany's generator, edit at your own risk! 
// If you do, be careful to not overwrite it when you run the generator again.
pragma Singleton

import QtQuick

QtObject {
    signal stateChanged()
    signal redoing(var scope, bool active)
    signal undoing(var scope, bool active)
}
